'use strict';

const debuglog = __debug.bind(null, __filename);

const _ = require('lodash');
const request = require('request');
const validator = require('validator');
const url = require('url');

const configUtil = require('../utils/config-util');
const httpUtil = require('../utils/http-util');
const oauth = require('../routes/oauth2');
const manifest = require('../manifest');

const TRIM_LENGTH = 100;

const MESSAGES= {
  network_error: 'Error in establishing connection',
  template_error: 'Error while substituting the templates',
  invalid_content_type: 'Unsupported content type'
};
const SUPPORTED_ENCODINGS = ['base64'];
const DEFAULT_ENCODING = 'base64';
const VALID_CONTENT_TYPES = [
  'application/json',
  'application/xml',
  'text/html',
  'text/xml',
  'application/jsonp',
  'text/plain',
  'application/vnd.api+json'
];

const utility = {
  encode: (data, encoding) => {
    encoding = SUPPORTED_ENCODINGS.includes(encoding) ? encoding : DEFAULT_ENCODING;
    return new Buffer(data).toString(encoding);
  }
};

function isValidContentType(response) {
  const headers = response.headers;
  const contentType = headers['content-type'] ? headers['content-type'].split(';')[0] : null;

  debuglog(`Looking for "${contentType}" in ${VALID_CONTENT_TYPES}`);

  if (contentType && !VALID_CONTENT_TYPES.includes(contentType)) {
    return false;
  }

  return true;
}

function isWhitelistedDomain(whitelistedDomains, url) {
  for (let domain of whitelistedDomains) {
    if (domain.startsWith('https://*')) {
      domain = domain.replace('*', '[a-z0-9-]*');
      const exp = new RegExp(domain);

      if (url.match(exp)) {
        debuglog(`Matched "${url}" with ${domain}`);
        return true;
      }
    }
    else if (domain === url) {
      debuglog(`Matched "${url}" with ${domain}`);
      return true;
    }
  }

  return false;
}

function isValidDomain(whitelistedDomains, endpoint) {
  const urlInfo = url.parse(endpoint);

  debuglog(`Parsed ${endpoint} as ${JSON.stringify(urlInfo)}`);

  if (urlInfo.protocol === 'https:' && validator.isFQDN(urlInfo.hostname)) {
    return isWhitelistedDomain(whitelistedDomains, endpoint);
  }

  debuglog(`Domain ${endpoint} is not valid.`);

  return false;
}

function fetchOauthConfigs(req) {
  const oauthCredential = oauth.fetchCredentials(req) || {};
  const oauthIparams = oauth.fetchOauthIparams();

  return {
    access_token : oauthCredential.access_token,
    oauth_iparams : oauthIparams || {}
  };
}

function substituteTemplates(options, templates) {
  debuglog(`Substituing Templates ${JSON.stringify(options)} with ${JSON.stringify(templates)}`);

  return JSON.parse(_.template(JSON.stringify(options))(templates));
}

module.exports = {
  execute: (req, res) => {
    let templates = {
      iparam: configUtil.getValuesForLocalTesting(),
      oauth_iparams: {}
    };
    let reqOptions = req.body.data;

    if ((_.includes(manifest.features, 'agent_oauth') ||
      _.includes(manifest.features, 'oauth')) &&
      req.body.data.isOAuth) {

      const oauthTokens = fetchOauthConfigs(req);

      templates = _.extend(templates, oauthTokens);
      debuglog(`Extending template values with oauth tokens and params ${JSON.stringify(oauthTokens)}`);
    }

    try {
      reqOptions = substituteTemplates(reqOptions, Object.assign({
        encode: utility.encode
      }, templates));

      const urlObject = url.parse(reqOptions.url);
      const domain = `${urlObject.protocol}//${urlObject.hostname}`;

      if (!isValidDomain(substituteTemplates(manifest.whitelistedDomains, templates), domain)) {
        return res.status(httpUtil.status.ok).send({
          status: httpUtil.status.bad_request,
          headers: {},
          response: 'This domain has not been whitelisted',
          errorSource: APP_SOURCE
        });
      }
    }
    catch (err) {
      debuglog(`Proxy errored with ${err.message}`);
      return res.status(httpUtil.status.ok).send({
        status: httpUtil.status.bad_request,
        headers: {},
        response: MESSAGES.template_error,
        errorSource: APP_SOURCE
      });
    }

    debuglog(`Making proxy call with options as ${JSON.stringify(reqOptions)}`);

    return request(reqOptions, function(error, response, responseBody) {
      debuglog(`Proxy came back with error as ${error && error.message}, body as ${(JSON.stringify(responseBody) || '').slice(0, TRIM_LENGTH)}`);

      if (error) {
        return res.status(httpUtil.status.ok).send({
          status: httpUtil.status.bad_gateway,
          headers: {},
          response: MESSAGES.network_error,
          errorSource: APP_SOURCE
        });
      }

      if (!isValidContentType(response)) {
        return res.status(httpUtil.status.ok).send({
          status: httpUtil.status.unsupported_media,
          headers: {},
          response: MESSAGES.invalid_content_type,
          errorSource: APP_SOURCE
        });
      }

      const responsePayload = {
        status: response.statusCode,
        headers: response.headers,
        response: responseBody
      };

      if (response.statusCode >= httpUtil.status.bad_request) {
        responsePayload.errorSource = APP_SOURCE;
      }

      res.status(httpUtil.status.ok).send(responsePayload);
    });
  }
};
