'use strict';

const debuglog = __debug.bind(null, __filename);

const _ = require('lodash');
const archiver = require('archiver');
const async = require('async');
const fs = require('fs-extra');
const path = require('path');

const eh = require('../utils/err');
const fileUtil = require('../utils/file-util');
const validator = require('./validate');
const digestFile = require('../utils/digest-file-handler');
const manifest = require('../manifest');
const nsUtil = require('../utils/ns-resolver');
const validationConst = require('../validations/constants').validationContants;
const writeMetric = require('../utils/write-metric');
const productInfo = require('../utils/product-info-util');

const zip = archiver('zip');
const cwd = process.cwd();
const pkgName = `${nsUtil.getRootFolder()}${nsUtil.pkgExt}`;
const PACKAGE_ERROR_MESSAGE = 'Error while generating package.';
const SERVER_DIR_ALLOWED_FILES = /.js$/;
const SERVER_DIR_EXCLUDED_FOLDERS = ['node_modules', 'etc', 'test_data'];
const DOCUMENTATION_REFERENCE = productInfo.getDeployLink();

function addFilesToZip() {

  async.series([

    function addAppDirectory(callback) {
      if (!fs.existsSync('./app')) {
        debuglog('Cannot find app folder while packing.');
        return callback();
      }
      fileUtil.getFiles({dir: './app'}, function(err, files) {
        if (err) {
          debuglog(`Error while adding app directory ${err.message}`);
          eh.error(new Error(PACKAGE_ERROR_MESSAGE));
        }
        files.forEach((file) => {
          debuglog(`Packing file ./app/${file}`);
          zip.file(`./${file}`);
        });
        callback(null);
      });
    },

    function addBackendDirectory(callback) {
      if (!_.includes(manifest.features, 'backend')) {
        debuglog('Not a backend app, skipped looking for server folder.');
        return callback();
      }
      async.each(_.difference(fs.readdirSync('./server'), SERVER_DIR_EXCLUDED_FOLDERS), function(file, cb) {
        const serverChild = `./server/${file}`;

        if (fs.lstatSync(serverChild).isDirectory()) {
          fileUtil.getFiles({
            dir: serverChild,
            match: SERVER_DIR_ALLOWED_FILES
          }, function(err, files) {
            if (err) {
              eh.error(new Error(PACKAGE_ERROR_MESSAGE));
            }
            files.forEach((file) => {
              debuglog(`Packing file ${serverChild}/${file}`);
              zip.file(`./${file}`);
            });
            cb();
          });
        }
        else {
          debuglog(`Packing file ${serverChild}`);
          zip.file(serverChild);
          cb();
        }
      }, function() {
        callback();
      });
    },

    function addConfigDirectory(callback) {
      fileUtil.getFiles({dir: './config'}, function(err, files) {
        if (err) {
          debuglog(`Error while adding config directory ${err.message}`);
          eh.error(new Error(PACKAGE_ERROR_MESSAGE));
        }
        files.forEach((file) => {
          debuglog(`Packing file config/${file}.`);
          zip.file(`./${file}`);
        });
        callback(null);
      });
    },

    function addManifest(callback) {
      zip.file('./manifest.json');
      debuglog('Packing manifest.');
      callback(null);
    },

    function addMetrics(callback) {
      writeMetric.addMetrics( function() {
        callback(null);
      });
    },

    function addActionFile(callback) {
      if (fileUtil.fileExists('./actions.json')) {
        zip.file('./actions.json');
        debuglog('Packing Actions.');
      }
      callback(null);
    },

    function addReport(callback) {
      if (fileUtil.fileExists('./.report.json')) {
        zip.file('./.report.json');
        debuglog('Packing report.');
      }
      callback(null);
    },

    function addDigest(callback) {
      digestFile.genDigestFile(function() {
        zip.file('./digest.md5');
        debuglog('Packing digest.');
        callback(null);
      });
    }

  ], function() {
    debuglog('Finalizing zip.');
    zip.finalize();
  });
}

function cleanFiles(err) {
  digestFile.delDigestFile();
  writeMetric.removeMetrics();
  if (err) {
    debuglog('Deleting dist/ folder.');
    fileUtil.deleteFile(`${cwd}/dist`);
  }
}

function zipPackage(output, cb) {
  output.on('close', function(err) {
    if (err) {
      cleanFiles(true);
      debuglog(`Error which creating package ${err.message}`);
      eh.error(new Error(PACKAGE_ERROR_MESSAGE));
    }

    const postPkgValidation = validator.run(validationConst.POST_PKG_VALIDATION);
    const product = Object.keys(manifest.product)[0];

    eh.printError('Packing failed due to the following issue(s):', postPkgValidation);
    console.log(`App packed successfully at ${path.sep}dist${path.sep}${pkgName}`);
    console.log(`Upload this file to the marketplace by following the instructions at ${DOCUMENTATION_REFERENCE[product]}`);
    cleanFiles();

    if (typeof cb === 'function') {
      cb();
    }
  });

  zip.on('error', function() {
    cleanFiles(true);
    eh.error(new Error(PACKAGE_ERROR_MESSAGE));
  });

  zip.pipe(output);

  addFilesToZip(zip);
}

function packApp(cb) {
  const dest = cwd + `/dist/${pkgName}`;

  fileUtil.ensureFile(dest);
  const output = fs.createWriteStream(dest);

  zipPackage(output, cb);
}

module.exports = {
  run: function(cb) {
    debuglog(`'pack' called with ${JSON.stringify(arguments)}`);

    var errors = validator.run(validationConst.PRE_PKG_VALIDATION);

    if (errors.length > 0) {
      eh.printError('Packing failed due to the following issue(s):', errors);
    }

    try {
      packApp(cb);
    }
    catch (e) {
      eh.printError('Some exception happened during pack - ', e);
    }
  }
};
