'use strict';

const debuglog = __debug.bind(null, __filename);

const _ = require('lodash');
const bodyParser = require('body-parser');
const express = require('express');
const fs = require('fs');
const path = require('path');
const ws = require('express-ws');

const beEvents = require('../routes/beevents');
const configs = require('../web/routes/configs.js');
const coverageUtil = require('../utils/coverage.js');
const dependencyInstaller = require('../utils/install');
const dpRouter = require('../routes/data-pipe');
const eh = require('../utils/err');
const fileUtil = require('../utils/file-util');
const httpUtil = require('../utils/http-util');
const iframe = require('../routes/iframe');
const manifest = require('../manifest');
const oauth2 = require('../routes/oauth2');
const util = require('../utils/helper');
const validator = require('./validate');
const validationConst = require('../validations/constants').validationContants;
const webEvents = require('../web/routes/events.js');
const watcher = require('../watcher');
const actions = require('../web/routes/actions.js');

const app = express();
const expressWs = ws(app);
const istanbulBodyParser = bodyParser.json({
  limit: '5MB'
});
const normalBodyParser = bodyParser.json();
const staticMiddleware = express.static('./app');
const exWss = expressWs.getWss('/notify-change');
const productInfo = require('../utils/product-info-util');

const DOMAIN = productInfo.getTestLink();
const HTTP_PORT = 10001;
const WS_CLOSE_TIMEOUT = 4000;
const WSS_SCRIPT = `
  <script>
    var ws = new WebSocket('ws://localhost:10001/notify-change');
    ws.onmessage = function(e) {
      location.reload();
    };
  </script>
`;

global.PLATFORM_SOURCE = 'PLATFORM';
global.APP_SOURCE = 'APP';

function logServerStartMsg() {
  const product = Object.keys(manifest.product)[0];

  console.log(`Starting local testing server at http://*:${HTTP_PORT}/\
    \nAppend 'dev=true' to your ${util.capitalize(product)} account URL to start testing\
    \ne.g. ${DOMAIN[product]}\
    \nQuit the server with Control-C.`);
  if (_.includes(manifest.features, 'backend')) {
    console.log('\nTo simulate product, app setup, and external events, visit - http://localhost:10001/web/test');
  }
  if (fileUtil.fileExists(`${process.cwd()}/config/iparams.html`)) {
    console.log('To test the custom installation page, visit - http://localhost:10001/custom_configs');
  }
}

module.exports = {
  run: (clearCoverage, skipCoverage) => {
    const validationMessages = validator.run(validationConst.RUN_VALIDATION);

    eh.printError('The local server could not be started due to the following issue(s):', validationMessages);

    if (clearCoverage) {
      coverageUtil.dispose();
    }

    if (skipCoverage) {
      coverageUtil.snooze();
    }

    app.set('views', path.join(__dirname + '/../web/views'));

    /**
     *  Istanbul might send back massive coverage reports as JSON.
     *  So have a body-parser with higher limits for that route
     *  alone.
     */

    app.use((req, res, next) => {
      if (req.path === '/iframe/coverage') {
        return istanbulBodyParser(req, res, next);
      }

      return normalBodyParser(req, res, next);
    });

    app.use('/web/assets', express.static(__dirname + '/../web/assets'));

    app.post('/iframe/coverage', (req, res) => {
      const coverageStats = req.body;

      coverageUtil.update(coverageStats);

      res.send('OK');
    });

    app.use('/iframe/oauth', express.static(__dirname +'/../web/pages/oauth'));

    /*
    IFrame:
    All the contents inside the app directory being served as static files.
    */

    app.use('/iframe', (req, res, next) => {
      const fullPath = `./app${req.path}`;

      if (req.method === 'GET') {
        if (req.path.endsWith('.js')) {
          debuglog(`Responding with contents of ${req.path}`);
          return res.send(coverageUtil.instrument(fullPath));
        }

        if (req.path.endsWith('.html')) {
          debuglog(`Responding with contents of ${req.path}`);
          return res.send(fs.readFileSync(fullPath, 'utf8') + WSS_SCRIPT);
        }
      }

      return staticMiddleware(req, res, next);
    });

    app.use((req, res, next) => {
      httpUtil.enableCORS(res);
      next();
    });


    app.ws('/notify-change', (ws) => {
      if (process.env.NODE_ENV === 'test') {
        setTimeout(() => {
          ws.close();
        }, WS_CLOSE_TIMEOUT);
      }
    });

    watcher.watch((data) => {
      exWss.clients.forEach((client) => {
        client.send(data);
      });
    });

    dpRouter.run(app);
    iframe.run(app);
    configs.run(app);
    beEvents.run(app);
    webEvents.run(app);
    oauth2.run(app);
    actions.run(app);

    // Finally, listen:
    const server = app.listen(HTTP_PORT, () => {
      if (_.includes(manifest.features, 'backend')) {
        dependencyInstaller.run((err) => {
          if (err) {
            console.log(err);
            process.exit(1);
          }
          logServerStartMsg();
        });
      } else {
        logServerStartMsg();
      }
    });

    process.on('uncaughtException', (err) => {
      if (err.code === 'EADDRINUSE') {
        console.error('Another instance of server running? Port in use.');
      }
      else {
        console.error(err);
      }
      process.exit(1);
    });

    return server;
  }
};
