'use strict';

const _ = require('lodash');
const jQueryDeferred = require('jquery-deferred');
const request = require('request');

const configUtil = require('../../utils/config-util');
const httpUtil = require('../../utils/http-util');
const oauth = require('../../routes/oauth2');
const manifest = require('../../manifest');

const DP_ROUTER_URL = 'http://localhost:10001/dprouter';

const MESSAGES = {
  internal_server_error: 'Internal server error',
  network_error: 'Error in establishing connection',
  timeout_error: 'Timeout error while processing the request',
  template_error: 'Error while substituting the templates',
  invalid_content: 'Unsupported content type'
};
const VALID_CONTENT_TYPES = [
  'application/json',
  'application/xml',
  'text/html',
  'text/xml',
  'application/jsonp',
  'text/plain',
  'application/vnd.api+json'
];
const APP_SOURCE = 'APP';
const PLATFORM_SOURCE = 'PLATFORM';
const TIMEOUT_RESP = {
  status: httpUtil.status.gateway_timeout,
  headers: {},
  response: MESSAGES.timeout_error,
  errorSource: APP_SOURCE
};
const TEMPLATE_ERR_RESP = {
  status: httpUtil.status.bad_request,
  headers: {},
  response: MESSAGES.template_error,
  errorSource: APP_SOURCE
};
const NETWORK_ERR_RESP = {
  status: httpUtil.status.bad_gateway,
  headers: {},
  response: MESSAGES.network_error,
  errorSource: APP_SOURCE
};
const INVALID_CONTENT_RESP = {
  status: httpUtil.status.unsupported_media,
  headers: {},
  response: MESSAGES.invalid_content,
  errorSource: APP_SOURCE
};
const TIMEOUT_CODE = 'ETIMEDOUT';
const SOCKET_TIMEOUT = 'ESOCKETTIMEDOUT';
const REQUEST_METHODS = [
  'get',
  'post',
  'put',
  'patch',
  'delete'
];
const REFRESH_FAILURE_STATUS = 210;
const REQUEST_TIMEOUT = 4500;
const REDIRECTS_FOLLOW = false;
const SUPPORTED_ENCODINGS = ['base64'];
const DEFAULT_ENCODING = 'base64';

const utility = {
  encode: (data, encoding) => {
    encoding = _.includes(SUPPORTED_ENCODINGS, encoding) ? encoding: DEFAULT_ENCODING;
    return new Buffer(data).toString(encoding);
  }
};

class Request {
  constructor() {
    // eslint-disable-next-line new-cap
    this.requestDefer = jQueryDeferred.Deferred();
  }

  makeRequest(action, url, options, recursive) {
    const self = this;

    const getOauthCredential = () => {
      if (!options.isOAuth) {
        // Sending empty object here for whitelist domain substitution.
        return { oauth_iparams: {} };
      }
      const oauthCredential = _.includes(manifest.features, 'oauth') ? oauth.fetchCredentials() : {};

      return {
        access_token: oauthCredential['access_token'] || '',
        oauth_iparams: oauthCredential['oauth_iparams'] || {}
      };
    };
    const params = _.extend({
      url: url,
      method: action,
      followAllRedirects: REDIRECTS_FOLLOW,
      timeout: REQUEST_TIMEOUT
    }, options);

    const templates = _.extend({
      iparam: configUtil.getValuesForLocalTesting(),
      encode: utility.encode
    }, getOauthCredential());

    try {
      const modifiedParams = JSON.parse(_.template(JSON.stringify(params))(templates));

      request(modifiedParams, (err, response, body) => {
        if (err) {
          const errResp = (err.code === TIMEOUT_CODE || err.code === SOCKET_TIMEOUT)
            ? TIMEOUT_RESP : NETWORK_ERR_RESP;

          self.requestDefer.reject(errResp);
        }
        else {
          const contentType = (response.headers['content-type']) ? response.headers['content-type'].split(';')[0] : null;

          if (response.statusCode === httpUtil.status.unauthorized &&
            params.isOAuth && !recursive) {
            request({
              method: 'POST',
              url: DP_ROUTER_URL,
              headers: { 'mkp-route' : 'oauth'},
              json: { action: 'refresh'}
            }, (err, resp, body) => {
              if (err) {
                return self.requestDefer.reject(err);
              }
              if (body && (body.status === httpUtil.status.bad_request ||
                 body.status === REFRESH_FAILURE_STATUS)) {
                return self.requestDefer.reject(_.extend(body, {errorSource: PLATFORM_SOURCE}));
              }
              return self.makeRequest(action, url, options, true);
            });
          }
          else if (contentType && !_.includes(VALID_CONTENT_TYPES, contentType)) {
            return self.requestDefer.reject(INVALID_CONTENT_RESP);
          }
          else {
            const proxyResponse = {
              status: response.statusCode,
              headers: response.headers,
              response: body
            };

            if (response.statusCode >= httpUtil.status.bad_request) {
              _.extend(proxyResponse, { errorSource: APP_SOURCE });
              return self.requestDefer.reject(proxyResponse);
            }
            self.requestDefer.resolve(proxyResponse);
          }
        }
      });
    } catch (err) {
      self.requestDefer.reject(TEMPLATE_ERR_RESP);
    }
    return self.requestDefer.promise();
  }
}

class RequestApi {
  constructor(args) {
    const self = this;

    REQUEST_METHODS.forEach((method) => {
      self[method] = (url, options) => {
        return new Request(args).makeRequest(method, url, options);
      };
    });
  }
}

module.exports = RequestApi;
