'use strict';

/* eslint-disable no-underscore-dangle */

const fs = require('fs');
const vm = require('vm');
const _ = require('lodash');
const path = require('path');
const util = require('util');
const uuid = require('uuid');

const serverScriptPath = `${process.cwd()}/server/server.js`;

const coverageUtil = require('../utils/coverage');
const httpUtil = require('../utils/http-util');
const manifest = require('../manifest');
const RequestAPI = require('./features/request');
const ScheduleAPI = require('./features/schedule');

const INVALID_DATA_ERR_MSG = 'The error should be a JSON Object with a message or a status parameter.';
const TIMEOUT_ERR_MSG = 'Timeout error while processing the request.';
const FW_ERROR_MSG = 'Error while executing App server script!';
const FEATURE_DEFINITIONS = {
  db : '$db'
};

function createModuleError(modulePath) {
  const error = new Error(`Cannot find module "${modulePath}"`);

  error.code = 'MODULE_NOT_FOUND';
  return error;
}

function loadDependency(dpn) {
  if (dpn.startsWith('./') || dpn.startsWith('../')) {
    dpn = path.basename(dpn);
  }
  return require(`${process.cwd()}/server/node_modules/${dpn}`);
}

function getFeatures(self) {
  const enabledFeatures = {};

  _.map(_.intersection(manifest.features, _.keys(FEATURE_DEFINITIONS)), (featureName) => {
    enabledFeatures[FEATURE_DEFINITIONS[featureName]] = require(`./features/${featureName}`);
  });
  enabledFeatures.generateTargetUrl = (options) => {
    return require('./features/generateTargetUrl')(self, options);
  };
  enabledFeatures.$request = new RequestAPI();
  enabledFeatures.$schedule = new ScheduleAPI(self);
  return enabledFeatures;
}

const APP_EVENT_TIMEOUT = 10000,
  PRODUCT_EVENT_TIMEOUT = 20000,
  SMI_TIMEOUT = 5000;

class Framework {

  constructor(event, req, res) {
    const self = this;

    this.event = event;
    this.req = req;
    this.res = res;

    this.cachedModules = {};

    this.serverScript = coverageUtil.instrument(serverScriptPath);
    /**
      We are running the developers code in the sandbox.
      By running in sandbox, we allow the user to only use certain methods.
      We use require to load the node modules [in loadDependency] as it contains the methods, which are outside of the context.
      We load the local library files[ loadLib ] by reading the files, restricting the user from using only the methods passed to the context.
    */
    this.sandboxAPI = _.extend({
      __fdkcoverage__: {},
      exports: module.exports,
      console: {
        log: console.log,
        info: console.info,
        error: console.error
      },
      loadLib: util.deprecate(function loadLib(fileName) {
        fileName = path.basename(fileName);
        const libScript = coverageUtil.instrument(`${process.cwd()}/server/lib/${fileName}.js`);

        return self.sandboxExecutor(libScript);
      }, '`loadLib` will be deprecated soon. Please use `require`.'),
      loadDependency: util.deprecate(loadDependency, '`loadDependency` will be deprecated soon. Please use `require`.'),
      require:  (relativePath) => {
        if (relativePath.startsWith('./') || relativePath.startsWith('../')) {
          const nestedPath = [ `${process.cwd()}/server/` ];
          const basePath = path.resolve.apply(null, nestedPath);
          const fullPath = path.normalize(`${basePath}/${relativePath}.js`);

          if (self.cachedModules.hasOwnProperty(fullPath)) {
            return self.cachedModules[fullPath];
          }

          if (!fullPath.startsWith(path.resolve(`${process.cwd()}/server/`))) {
            throw createModuleError(relativePath);
          }

          if (!fs.existsSync(fullPath)) {
            throw createModuleError(relativePath);
          }

          const script = coverageUtil.instrument(fullPath);

          nestedPath.push(path.dirname(relativePath));
          const exported = self.sandboxExecutor(script);

          nestedPath.pop();
          self.cachedModules[fullPath] = exported;

          return exported;
        }

        return loadDependency(relativePath);
      }
    }, getFeatures(self, req));

    (() => {
      let previous;
      const INTERVAL_MS = 2000;

      setInterval(() => {
        const current = JSON.stringify(self.sandboxAPI.__fdkcoverage__);

        if (previous !== current) {
          coverageUtil.update(self.sandboxAPI.__fdkcoverage__);
          previous = current;
        }
      }, INTERVAL_MS).unref();
    })();

    // When the developer doesn't use renderData then we don't return the requestID in response. It would be an empty success response.
    _.extend(this.sandboxAPI, {
      renderData : (error, output) => {
        try {
          let returnVal = null;
          let statusCode = 200;
          const requestID = self.req.RequestId;
          let showInUI = false;

          if (error) {
            let response = {};

            if (_.isArray(error) || !(_.isObject(error)) ||
              (_.isUndefined(error.message) && !error.status)) {
              statusCode = httpUtil.status.bad_request;
              response = {
                status : statusCode,
                message : INVALID_DATA_ERR_MSG,
                errorSource: APP_SOURCE };
            }
            else {
              if (_.isUndefined(error.message)) {
                error.message = null;
              }
              statusCode = error.status || httpUtil.status.internal_server_error;
              response = {
                status : statusCode,
                message : error.message,
                errorSource: APP_SOURCE
              };
              showInUI = true;
            }
            returnVal = _.extend({
              requestID : requestID
            }, response );
          }
          else {
            returnVal = {
              requestID : requestID
            };
            if (!_.isUndefined(output)) {
              returnVal = _.extend({
                requestID : requestID
              }, {
                response : output
              });
            }
          }

          self.res.header('showinui', showInUI);
          self.res.status(statusCode).json(returnVal);
        }
        catch (e) {
          console.log(e);
        }
      }
    });
    this.sandboxAPI.global = this.sandboxAPI;

    // eslint-disable-next-line new-cap
    this.contextifiedGlobal = new vm.createContext(this.sandboxAPI);
  }

  getMethodName() {
    return this.event.categoryArgs.methodName;
  }

  getMethodArgs() {
    return this.event.categoryArgs.methodParams;
  }

  execute(methodName, args) {
    const jsonArgs = JSON.stringify(args || {});
    const executorScript = `__svrScript['${methodName}'](${jsonArgs})`;

    _.extend(this.contextifiedGlobal, { __svrScript: this.svrExecScript });
    this.sandboxExecutor(executorScript);
  }

  sandboxExecutor(script) {
    const serverScript = this.constructScript(script);

    return serverScript.runInContext(this.contextifiedGlobal, { timeout: this.event.timeout });
  }

  constructScript(script) {
    script = `(function() {var exports = {}; ${script}; return exports;})()`;
    return new vm.Script(script, {
      filename: 'server.js'
    });
  }

  handleError(err) {
    console.log(err);
    const output = {
      requestID : this.req.RequestId,
      status : httpUtil.status.internal_server_error,
      message : `${FW_ERROR_MSG} - ${err.message}`,
      errorSource: APP_SOURCE
    };

    this.res.status(httpUtil.status.ok).json(output);
  }

}

class Request extends Framework {

  constructor(event, req, res) {
    event.timeout = event.timeout || SMI_TIMEOUT;
    super(event, req, res);
  }

  execute() {
    try {
      const self = this;
      const TIMEOUT_MS = 5000;

      setTimeout(() => {
        if (!self.res._headerSent) {
          self.res.status(httpUtil.status.ok).send({
            status: httpUtil.status.gateway_timeout,
            message: TIMEOUT_ERR_MSG,
            errorSource: APP_SOURCE
          });
        }
      }, TIMEOUT_MS);
      super.execute(this.getMethodName(), this.getMethodArgs());
    }
    catch (err) {
      this.handleError(err);
    }
  }

}

class Event extends Framework {

  constructor(event, req, res) {
    super(event, req, res);
  }

  getMethodName() {
    let methodName = super.getMethodName();
    const events = this.svrExecScript.events || [];

    for (const i in events) {
      if (events[i].event === methodName) {
        methodName = events[i].callback;
        break;
      }
    }
    return methodName;
  }

  execute() {
    try {
      super.execute(this.getMethodName(), this.getMethodArgs());
    }
    catch (err) {
      this.handleError(err);
    }
  }
}

class ProductEvent extends Event {
  constructor(event, req, res) {
    event.timeout = event.timeout || PRODUCT_EVENT_TIMEOUT;
    super(event, req, res);
  }
}

class AppEvent extends Event {
  constructor(event, req, res) {
    event.timeout = event.timeout || APP_EVENT_TIMEOUT;
    super(event, req, res);
  }
}

function fetchEventObj(req, res) {
  const event = req.body;

  switch (req.body.categoryName) {
  case 'request':
    return new Request(event, req, res);
  case 'appEvent':
    return new AppEvent(event, req, res);
  case 'productEvent':
    return new ProductEvent(event, req, res);
  }
}

function handler (req, res) {
  req.RequestId = `${uuid.v4()}`;
  if (req.body.categoryName) {
    const obj = fetchEventObj(req, res);

    try {
      obj.svrExecScript = obj.sandboxExecutor(obj.serverScript);
      obj.execute();
    }
    catch (err) {
      console.log(err);
      const output = {
        requestID : req.RequestId,
        status : httpUtil.status.internal_server_error,
        message : `${FW_ERROR_MSG} - ${err.message}`,
        errorSource: APP_SOURCE
      };

      res.status(httpUtil.status.ok).json(output);
    }
  }
  else {
    const output = {
      requestID: req.RequestId,
      status: httpUtil.status.internal_server_error,
      message: FW_ERROR_MSG,
      errorSource: PLATFORM_SOURCE
    };

    res.status(httpUtil.status.ok).json(output);
  }
}

module.exports = handler;
