'use strict';

const _ = require('lodash');
const fs = require('fs');
const OAuth2Strategy = require('passport-oauth2').Strategy;
const passport = require('passport');
const refresh = require('passport-oauth2-refresh');
const uuid = require('uuid');

const configUtil = require('../utils/config-util');
const fileUtil = require('../utils/file-util');
const httpUtil = require('../utils/http-util');
const cryptoUtil = require('../utils/crypto-util');
const oauthUtil = require('../utils/oauth-util');

const charset = 'utf8';
const STRATEGY = 'oauth2';
const CACHED_VIEWS = {};
const BLANK_STATE = 'State is blank';
const INVALID_STATE = 'Invalid state';
const REFRESH_FAILURE_STATUS = 210;

let cachedOAuthConfig = null;
let token_type = null;

function getOAuthConfig() {
  if (cachedOAuthConfig) {

    return cachedOAuthConfig;
  }
  cachedOAuthConfig = fs.readFileSync('./config/oauth_config.json', charset);

  return cachedOAuthConfig;
}

function isNotAgent() {
  if (token_type) {

    return token_type !== 'agent';
  }
  token_type = JSON.parse(getOAuthConfig()).token_type;

  return token_type !== 'agent';
}

/**
  oauthIparams to be global to preserve the oauthIparams over re-direction
*/
let oauthIparams = {};
let callback = '';

function fetchOauthIparams(){

  return oauthUtil.fetchIparams();
}

function fetchCredentials(req) {
  if (isNotAgent()) {

    return oauthUtil.fetchCredentialsForAccount();
  }

  return oauthUtil.fetchCredentialsForAgent(req);
}

function interpolateJSON(oauthConfig, oauthIparams) {

  return _.template(oauthConfig)({
    oauth_iparams: oauthIparams
  });
}

function tokenHandler(accessToken, refreshToken, oauthIparams, res) {
  const credentials = {
    access_token : isNotAgent() ? accessToken : cryptoUtil.encryptToken(accessToken)
  };

  if (refreshToken) {
    Object.assign(credentials, {
      refresh_token: isNotAgent() ? refreshToken : cryptoUtil.encryptToken(refreshToken)
    });
  }

  if (isNotAgent()) {
    oauthUtil.storeIparams(oauthIparams);
    oauthUtil.storeCredentials(credentials);

    res.writeHead(httpUtil.status.found, {
      'Location': callback
    });

    return res.end();
  }

  return res.render('oauth_landing.html', {
    tokens: credentials
  });
}

function prepareOauthConfig(req) {
  /**
    If the authorization takes place for the first time, oauth_iparams are
    sent via query params from the form.

    Else use the stored oauth_iparams.
  */

  if (isNotAgent()) {
    const oauthIparamsCredentials = fetchOauthIparams();

    if (req.query.oauth_iparams) {
      oauthIparams = JSON.parse(req.query.oauth_iparams);
    }
    else if (oauthIparamsCredentials && oauthIparamsCredentials.oauth_iparams) {
      oauthIparams = oauthIparamsCredentials;
    }

    return JSON.parse(interpolateJSON(getOAuthConfig(), oauthIparams));
  }

  return JSON.parse(getOAuthConfig());

}

function oauthStrategyDetails(oauthConfig) {

  return {
    clientID: oauthConfig['client_id'],
    clientSecret: oauthConfig['client_secret'],
    callbackURL: 'http://localhost:10001/auth/callback',
    authorizationURL: oauthConfig['authorize_url'],
    tokenURL: oauthConfig['token_url']
  };
}

function oauthInit(req, res, next) {
  var oauthConfig = prepareOauthConfig(req);
  var strategyDetails = oauthStrategyDetails(oauthConfig);
  const oauthStrategy = new OAuth2Strategy(strategyDetails,
    function (accessToken, refreshToken) {
      tokenHandler(accessToken, refreshToken, oauthIparams, res);
    }.bind({
      'response': res
    })
  );

  if (!(_.isEmpty(oauthConfig.options))) {
    oauthStrategy.authorizationParams = () => {

      return oauthConfig.options;
    };
  }

  passport.use(oauthStrategy);
  refresh.use(oauthStrategy);
  if (next) {
    next();
  }
}

function refreshInit(req, res) {
  const data = fetchCredentials(req);

  if (data.refresh_token) {

    return refresh.requestNewAccessToken(STRATEGY, data['refresh_token'],
      (err, accessToken, refreshToken) => {
        const credentials = {
          access_token: accessToken || data['access_token'],
          refresh_token: refreshToken || data['refresh_token']
        };

        if (err) {

          return res.status(err.statusCode).send({
            status: err.statusCode,
            headers: {},
            response: err.data
          });
        }

        if (isNotAgent()) {
          oauthUtil.storeCredentials(credentials);
          if (accessToken || refreshToken) {

            return res.status(httpUtil.status.ok).json({
              success: true
            });
          }
        } else {
          const responsePayLoad = {
            action : 'refresh',
            tokens : {
              access_token : cryptoUtil.encryptToken(accessToken),
              refresh_token : cryptoUtil.encryptToken(refreshToken)
            }
          };

          return res.status(httpUtil.status.ok).send(responsePayLoad);
        }

      });
  }

  const responsePayLoad = {
    status: 210,
    headers: {},
    response: 'Access token cannot be refreshed'
  };

  return res.status(REFRESH_FAILURE_STATUS).send(responsePayLoad);
}

function verifyState(req, res, next) {
  const state = req.query.state;

  try {
    if (!state) {

      throw new Error(BLANK_STATE);
    }

    if (state !== oauthUtil.fetchState()) {

      throw new Error(INVALID_STATE);
    }
  }
  catch (e) {

    return res.end(INVALID_STATE);
  }

  return next();
}

module.exports = {

  fetchCredentials,

  fetchOauthIparams,

  isNotAgent,

  refresh: (req, res) => {
    oauthInit(req, res);
    refreshInit(req, res);
  },

  run: app => {
    app.engine('html', (templatePath, options, callback) => {
      if (templatePath in CACHED_VIEWS) {

        return callback(null, _.template(CACHED_VIEWS[templatePath])(options));
      }

      fs.readFile(templatePath, (err, templateContent) => {
        if (err) {

          return callback(err);
        }

        CACHED_VIEWS[templatePath] = templateContent;

        return callback(null, _.template(templateContent)(options));
      });
    });

    app.use('/auth/index', (req, res, next) => {
      if (req.query.callback) {
        callback = req.query.callback;
      }

      if (isNotAgent()){
        const oauthIParams = configUtil.getOAuthIparam().oauth_iparams;

        if (oauthIParams && !_.isEmpty(oauthIParams) && !req.query.oauth_bypass) {

          return res.send(fileUtil.readFile(__dirname + '/../web/pages/oauth-forms.html'));
        }
      }

      oauthInit(req, res, next);
    });

    app.use('/auth/callback', verifyState);

    app.use('/auth/callback', (req, res, next) => {
      oauthInit(req, res, next);
    });

    app.use('/accesstoken', (req, res, next) => {
      httpUtil.enableCORS(res);
      next();
    });

    app.get('/auth/index', (req, res) => {
      const oauthIparams = req.query.oauth_iparams ? JSON.parse(req.query.oauth_iparams) : {};
      const stateUuid = uuid.v4();

      oauthUtil.storeIparams(oauthIparams);
      oauthUtil.storeState(stateUuid);

      passport.authenticate(STRATEGY, {
        state: stateUuid
      })(req, res);
    });

    app.get('/auth/callback', passport.authenticate(STRATEGY, {
      session: false
    }));

    app.get('/accesstoken', (req, res) => {
      const credentials = fetchCredentials(req);

      httpUtil.enableCORS(res);

      if (credentials) {

        return res.send(credentials.access_token);
      }

      return res.status(httpUtil.status.not_found).send();
    });

    app.get('/oauth_iparams', (req, res) => {
      const oauthConfig = configUtil.getOAuthIparam();

      if (oauthConfig.oauth_iparams && !_.isEmpty(oauthConfig.oauth_iparams)) {

        return res.json(oauthConfig.oauth_iparams);
      }

      return res.json({});
    });
  }
};
