'use strict';

const _ = require('lodash');
const esprima = require('esprima');
const fs = require('fs');

const eventUtil = require('../utils/event-util');
const manifestConfigs = require('../manifest');
const validationConst = require('./constants').validationContants;

const debuglog = __debug.bind(null, __filename);
const VALID_KEYS = ['event', 'callback'];

function validateEvents(evtHash) {
  const errMsgs = [];
  const eventsArr = [];
  const events = evtHash.events;

  events.forEach(function(ev) {
    const diff = _.difference(_.keys(ev), VALID_KEYS);
    const eventName = ev.event;
    const callback = ev.callback;
    const product = Object.keys(manifestConfigs.product)[0];

    if (diff.length > 0) {
      errMsgs.push(`Invalid key in events array: '${diff}'`);
    }

    if (_.includes(eventsArr, eventName)) {
      errMsgs.push(`Event '${eventName}' is already mapped to a callback`);
    }

    if (!eventUtil.isValidEvent(product, eventName)) {
      errMsgs.push(`Invalid event: '${eventName}'`);
    }
    else {
      eventsArr.push(eventName);
    }
    if (!_.includes(evtHash.definedFunctions, callback)) {
      errMsgs.push(`Event callback '${callback}' is not defined.`);
    }
  });

  return errMsgs;
}

function getDefinedFunctions(hash) {
  const definedFunc = [];

  _.each(hash, function(prop) {
    if (prop.value.type === 'FunctionExpression') {
      definedFunc.push(prop.key.name);
    }
  });

  debuglog(`Found ${JSON.stringify(definedFunc)} as exported functions.`);

  return definedFunc;
}

function getEventsList() {
  const events = [];
  const code = fs.readFileSync( `${process.cwd()}/server/server.js`, 'utf8');
  const ast = esprima.parse(code).body;
  const exportsExp = _.find(ast, function(node) {
    return (node.type === 'ExpressionStatement'
      && node.expression.type === 'AssignmentExpression'
      && node.expression.left.name === 'exports');
  });

  let definedFunc = [];

  if (exportsExp) {
    const properties = exportsExp.expression.right.properties;

    definedFunc = getDefinedFunctions(properties);
    const eventProp = _.find(properties, function(prop) {
      return (prop.key.name === 'events'
        && prop.value.type === 'ArrayExpression');
    });

    const eventsHash = eventProp ? eventProp.value.elements : [];

    _.each(eventsHash, function(evt) {
      const values = evt.properties;
      const hash = {};

      _.each(values, function(p) {
        hash[p.key.name] = p.value.value;
      });
      events.push(hash);
    });
  }

  debuglog(`Found ${JSON.stringify(events)} as events.`);

  return {
    events: events,
    definedFunctions: definedFunc
  };
}

function isBackendApp() {
  return _.includes(manifestConfigs.features, 'backend');
}

module.exports = {
  validate() {
    try {
      let errMsgs = [];

      if (isBackendApp()) {
        const evtHash = getEventsList();

        errMsgs = validateEvents(evtHash);
      }
      return errMsgs;
    }
    catch (e) {
      debuglog(e);
      return [`Exception occured while validation: ${e.message}`];
    }
  },

  validationType: [validationConst.PRE_PKG_VALIDATION, validationConst.RUN_VALIDATION]
};
